c Program to search for spherical expanding shells in XYV datacubes using direct method.

c The program was created to search for spheric expanding shells in 
c     radio-interferometric XYV data-cubes. Without changes it can be used
c     to look in XYV data-cubes for any other compact features symmetric
c     relatively their centers. 


c  Requires FITSIO library to compile.

c   (c) Sergey Mashchenko 1999-2001

c  3.17 - direct analog of GSA code sf2.17.f.
c  3.18 - optional subtraction of background as a function of velocity.
c  3.19 - we get rid of smoothing.
c  3.20 - switched back to linear scale for Velocity.
c  3.21 - new option: dump. Dumps model images and rho images for extreme shells.
c  3.22 - background subtraction only for area around the shell. Also corrected
c          Vel scale.
c  3.23 - BS as linear function in X,Y coordinates.
c  3.24 - Extended output (file sf_ext.dat).
c  3.25 - Calculates mass using least squares fit.
c  3.26 - bigger mod??.fits file (the same size, az input FITS file).
c  3.27 - Extra output in *.dat file - shell integral brightness for each detection.
c  3.28 - Extra output in *.dat file - shell contrast.
c  3.29 - Fixed bug in cost_function - now works correctly with undefined data pixels.
           
c-------------------------------------------------------------------------
      program Shell_finder
      implicit NONE

      integer NX_max,NY_max,NV_max, MX_max,MY_max,MV_max, 
     #        LX_max,LY_max,LV_max, NR_max,NVel_max
c  Maximum data cube size:
      parameter (NX_max=111)
      parameter (NY_max=111)
      parameter (NV_max=111)
c  Maximum input Model image size:
      parameter (MX_max=55)
      parameter (MY_max=55)
      parameter (MV_max=55)
c  Maximum scaled Model size:
      parameter (LX_max=2*MX_max+1)
      parameter (LY_max=2*MY_max+1)
      parameter (LV_max=2*MV_max+1)

c  Maximum number of (Rad,Vel) combinations:
      parameter (NR_max=52)
      parameter (NVel_max=52)

      real Data(NX_max,NY_max,NV_max), Model(MX_max,MY_max,MV_max),
     #     Rpix(3), Delt(3), Rota(3), BG(NV_max),
     #     D, SFRAD,SFVEL, L0Model(LX_max,LY_max,LV_max),
     #     R1,R2,Vel1,Vel2, Sgm,Vsgm,
     #     LModel(LX_max,LY_max,LV_max), Center(3),Frac(3),dRl,dVl,
     #     Sum(LV_max),Sum_C(LV_max),VSum_C(LX_max,LY_max),
     #     VSum(LX_max,LY_max), Edat(6,NR_max,NVel_max)
     #     ,Data2(NX_max,NY_max,NV_max),Contr(NR_max,NVel_max)

      real*8 Rval(3)

      integer BlockSize, Bitpix, Naxis, Naxes(4), ID, IX,IY,IV,
     #        Laxes(4), Maxes(4),IR1,IR2,IV1,IV2,k,IR,IR0,IV0,
     #        strlen_trail, nch

      logical No_BUNIT, No_coord, BS, Dump

      character Type(3)*8, BUNIT*8, IName*256, MName*256
      
      equivalence (D,ID)

c-------------------------------------------------------------------------

      write (*,*) '  Shell finder (ver. 3.29)'
      write (*,*)
      
      write (*,*) ' Name of input data file:'
      read (*,'(A)') IName
      write (*,'(A)') IName
      write (*,*) ' Name of input model file:'
      read (*,'(A)') MName
      write (*,'(A)') MName
      write (*,*) ' X0 and maximum dX/Rad:'
      read (*,*) Center(1), Frac(1)
      write (*,*) Center(1), Frac(1)
      write (*,*) ' Y0 and maximum dY/Rad:'
      read (*,*) Center(2), Frac(2)
      write (*,*) Center(2), Frac(2)
      write (*,*) ' V0 and maximum dV/Vel:'
      read (*,*) Center(3), Frac(3)
      write (*,*) Center(3), Frac(3)
      write (*,*) ' Range for bubble radius:'
      read (*,*) R1,R2
      write (*,*) R1,R2
      write (*,*) ' Range for bubble velocity:'
      read (*,*) Vel1,Vel2
      write (*,*) Vel1,Vel2
      write (*,*) ' Steps for Radius and Velocity:'
      read (*,*) dRl, dVl
      write (*,*) dRl, dVl
      write (*,*) ' Perform background subtraction?'
      read (*,*) BS
      write (*,*) BS
      write (*,*) ' Dump?'
      read (*,*) Dump
      write (*,*) Dump
      

c  Reading the input model file:
      call Reading_input_image (MName,Naxis,Maxes,No_BUNIT,BUNIT,
     #         Type,Rval,Rpix,Delt,Rota,No_coord,
     #       Model,MX_max,MY_max,MV_max, BlockSize,Bitpix, SFRAD,SFVEL)

c  Reading the input data file:
      call Reading_input_image (IName,Naxis,Naxes,No_BUNIT,BUNIT,
     #         Type,Rval,Rpix,Delt,Rota,No_coord,
     #         Data,NX_max,NY_max,NV_max, BlockSize,Bitpix, SFRAD,SFVEL)

c  Magic blanking for spheric boundaries:
      ID = -1
      do IX=1,Maxes(1)
         do IY=1,Maxes(2)
            do IV=1,Maxes(3)
               if ((((IX-1))**2 + (IY-1)**2)/SFRAD**2 + 
     #           ((IV-1)/SFVEL)**2.gt.1.0 .and. Model(IX,IY,IV).eq.0.0) 
     #           Model(IX,IY,IV) = D
               end do
            end do
         end do


      call Direct_method (Data,NX_max,
     #       NY_max,NV_max,Naxes, LX_max,LY_max,LV_max,Laxes,
     #       Model,LModel,MX_max,MY_max,MV_max,Maxes,
     #       SFRAD,SFVEL,R1,R2,Vel1,Vel2, L0Model,
     #       Center,Frac,dRl,dVl,BS,BG,Sgm,Vsgm,Sum,Sum_C,VSum,VSum_C,
     #       Data2,Dump,Type,No_coord,BlockSize,Bitpix,Rval,Rpix,Delt,
     #       Rota,Edat,Contr,NR_max,NVel_max,IR1,IR2,IV1,IV2)

      if (BS)  then
         call Writing_output_image(Data,NX_max,NY_max,NV_max,
     #         'BS.fits', BlockSize, Naxes, BUNIT,No_BUNIT, Type,Rval,
     #         Rpix,Delt,Rota,No_coord,Bitpix)
         end if

         nch = strlen_trail(IName)
         open (3,file=IName(1:nch-5)//'.dat')
         write (3,'(A)') IName
         write (3,'(2(3E14.6,X))') (Center(k),k=1,3), (Frac(k),k=1,3)
         write (3,*) BS
         write (3,'(4I4,6E14.6)') IR1,IR2,IV1,IV2,R1,R2,dRl, Vel1,Vel2,
     #      dVl
         write (3,'(9E16.8)') (Rval(k),Rpix(k),Delt(k), k=1,3)
         do IR=IR1,IR2
            IR0 = IR-IR1+1
            do IV=IV1,IV2
               IV0 = IV-IV1+1
               write (3,'(2(1PE14.6),3I4,2(1PE14.6))') Edat(1,IR0,IV0),
     #              Edat(2,IR0,IV0),
     #              (int(Edat(k,IR0,IV0)),k=3,5),Edat(6,IR0,IV0),
     #              Contr(IR0,IV0)
               end do
            end do
         close(3)

      stop
      end
c-------------------------------------------------------------------------


      subroutine Direct_method (Data,NX_max,
     #       NY_max,NV_max,Naxes, LX_max,LY_max,LV_max,Laxes,
     #       Model,LModel,MX_max,MY_max,MV_max,Maxes,
     #       SFRAD,SFVEL,R1,R2,Vel1,Vel2,L0Model,
     #       Center,Frac,dRl,dVl,BS,BG,Sgm,Vsgm,Sum,Sum_C,VSum,VSum_C,
     #       Data2,Dump,Type,No_coord,BlockSize,Bitpix,Rval,Rpix,Delt,
     #       Rota,Edat,Contr,NR_max,NVel_max,IR1,IR2,IV1,IV2)

c  Calculation of cross-correlation coefficient by direct method.

      implicit NONE
      integer NX_max,NY_max,NV_max, LX_max,LY_max,LV_max,
     #        MX_max,MY_max,MV_max, NEM_max, NR_max,NVel_max
      parameter (NEM_max = 100)
      real Data(NX_max,NY_max,NV_max), St,St2,E,Cost_function,
     #     SFRAD,SFVEL, Limits(2,5), BG(NV_max),
     #     Model(MX_max,MY_max,MV_max),R1,R2,Vel1,Vel2, MXY,MV
     #     ,LModel(LX_max,LY_max,LV_max),
     #     Xmin(5), Rad,Vel,L0Model(LX_max,LY_max,LV_max),
     #     Center(3),Frac(3),dRl,dVl,Emin,Xk(5),Sgm,Vsgm,
     #     Sum(LV_max),Sum_C(LV_max),VSum_C(LX_max,LY_max),
     #     Xf2,Yf2,Vf2,VSum(LX_max,LY_max),Data2(NX_max,NY_max,NV_max),
     #     Rpix(3),Delt(3),Rota(3),D,RR2,AAA(3,3),be(3),
     #     Edat(6,NR_max,NVel_max), E2min, X2min(5),CM,Mass,
     #     M2min,Mmin,Bl,Contr(NR_max,NVel_max),Contr0,Contr2
      real*8 Rval(3)
      integer Naxes(4), Laxes(4), IR,IV,IR1,IR2,IR0,IV0,
     #        Maxes(4), Nt,  MMM, dIR,dIV,LXc,LYc,LVc,LX,LY,LV,
     #        IV1,IV2,JX,JY,JV,JX1,JX2,JY1,JY2,JV1,JV2,IX,IY,
     #        BlockSize, Bitpix,ID,digR,digV,NN2,k1,k2,k,IBl
      logical Wrong, BS, Dump, No_BUNIT, No_coord
      character Type(3)*8, BUNIT*8, No*2
      equivalence (D,ID), (Bl,IBl)

      IBl = -1

c  We subtract background from ellipsoid volume defined by R2,V2.
      if (BS)  then
c  Background as a function of velocity:
         do IV=1,Naxes(3)
            NN2 = 0
            do k1=1,3
               be(k1) = 0.0
               do k2=1,3
                  AAA(k1,k2) = 0.0
                  end do
               end do
            do IX=1,Naxes(1)
               do IY=1,Naxes(2)
c  3.22: smaller area for background subtraction:
                  RR2 =  ((IX-Center(1))/(R2*(1.2+Frac(1))))**2 +
     #                   ((IY-Center(2))/(R2*(1.2+Frac(2))))**2
                  D = Data(IX,IY,IV)
c  3.23: linear interpolation of background:
                  if (RR2.le.1.0 .and. ID.ne.-1) then
                     NN2 = NN2 + 1
                     AAA(1,1) = AAA(1,1)+IX**2
                     AAA(1,2) = AAA(1,2)+IX*IY
                     AAA(1,3) = AAA(1,3)+IX
                     AAA(2,2) = AAA(2,2)+IY**2
                     AAA(2,3) = AAA(2,3)+IY
                     be(1) = be(1) + IX*D
                     be(2) = be(2) + IY*D
                     be(3) = be(3) + D
                     else
c  Setting outskirts to Blank value:
                     Data(IX,IY,IV) = Bl
                     end if
                  end do
               end do
            AAA(3,3) = NN2
            AAA(2,1) = AAA(1,2)
            AAA(3,1) = AAA(1,3)
            AAA(3,2) = AAA(2,3)
            call gaussj (AAA,3,3,be,1,1)
c  Subtraction of background:
            do IX=1,Naxes(1)
               do IY=1,Naxes(2)
                  D = Data(IX,IY,IV)
                  RR2 =  ((IX-Center(1))/(R2*(1.2+Frac(1))))**2 +
     #                   ((IY-Center(2))/(R2*(1.2+Frac(2))))**2
                  if (RR2.le.1.0 .and. ID.ne.-1) then
                     Data(IX,IY,IV) = D-be(1)*IX-be(2)*IY-be(3)
                     end if
                  end do
               end do
            end do
         end if
         
         
      Emin = 1.e33
      ID = -1
      MMM = 0

c  Logarithmic scale for Rad, linear scale for Vel:
      IR1 = 0
c      IV1 = anint(Vel1/dVl)
      IV1 = 0
      IR2 = anint(alog(R2/R1)/dRl)
      IV2 = anint((Vel2-Vel1)/dVl)

      if (IR2-IR1+1.gt.NR_max .or. IV2-IV1+1.gt.NVel_max)  then
         write (*,*) ' Too small NR_max, NVel_max !'
         stop
         end if

c  Step for IR and IV - used for "dump" mode:
      if (Dump)  then
         dIR = IR2-IR1
         if (dIR .eq. 0)  dIR = 1
         dIV = IV2-IV1
         if (dIV .eq. 0)  dIV = 1
        else
         dIR = 1
         dIV = 1
        end if

      do IR=IR1,IR2,dIR
         Rad = R1 * exp(IR*dRl)
         Xk(4) = Rad
         write (*,'(2(A,I3),A)')  ' IR=',IR-IR1+1,' (maximum ',
     #                            IR2-IR1+1,')'
         do IV=IV1,IV2,dIV
            Vel = Vel1 + IV*dVl
            Xk(5) = Vel
            E2min = 1e33
c  Spatial boundaries for given Rad and Vel:
            call Calculate_limits (Rad,Vel,SFRAD,SFVEL,
     #           Naxes,Maxes, Laxes,Limits,MXY,MV,Wrong,
     #           Center,Frac)
c  Generating a model for given Rad and Vel. The model is stored in 3D
c    array LModel.
            call Generate_model (Model,MX_max,MY_max,MV_max,Maxes,
     #           LModel,LX_max,LY_max,LV_max,Laxes,
     #           MXY,MV,St,St2,Nt,CM)
c  Dumping the model image in Dump mode:
            if (Dump)  then
               digR = (IR-IR1)/dIR
               digV = (IV-IV1)/dIV
               write (No,'(I2.2)') digR*10 + digV
               do JX=1,Naxes(1)
                  do JY=1,Naxes(2)
                     do JV=1,Naxes(3)
                        Data2(JX,JY,JV) = D                        
                        end do
                     end do
                  end do
               end if
            JX1 = anint(Limits(1,1)*Rad+Center(1))
            JX2 = anint(Limits(2,1)*Rad+Center(1))
            JY1 = anint(Limits(1,2)*Rad+Center(2))
            JY2 = anint(Limits(2,2)*Rad+Center(2))
            JV1 = anint(Limits(1,3)*Vel+Center(3))
            JV2 = anint(Limits(2,3)*Vel+Center(3))
            do JX=JX1,JX2
               Xk(1) = JX
               if (Frac(1).ne.0.0)  then
                  Xf2 = ((JX-Center(1))/
     #                 nint(Frac(1)*Rad))**2
                  elseif (JX.ne.Center(1))  then
                     Xf2 = 1.e30
                    else
                     Xf2 = 0.0
                    end if
               do JY=JY1,JY2
                  Xk(2) = JY
                  if (Frac(2).ne.0.0)  then
                     Yf2 = ((JY-Center(2))/
     #                    nint(Frac(2)*Rad))**2
                    elseif (JY.ne.Center(2))  then
                     Yf2 = 1.e30
                    else
                     Yf2 = 0.0
                    end if
                  do JV=JV1,JV2
                     Xk(3) = JV
                     if (Frac(3).ne.0.0)  then
                        Vf2 = ((JV-Center(3))/
     #                       nint(Frac(3)*Vel))**2
                       elseif (JV.ne.Center(3))  then
                        Vf2 = 1.e30
                       else
                        Vf2 = 0.0
                       end if
c  Looking for a shell with a center within ellipsoid defined by Frac:
                    if (Xf2+Yf2+Vf2 .le. 1.0)  then
                     MMM = MMM + 1
                     E = Cost_function (Xk, Data,NX_max,NY_max,NV_max,
     #                    Naxes,
     #                    LModel,LX_max,LY_max,LV_max,Laxes,
     #                    St,St2,Nt,Center,CM,Mass,Contr0)
                     if (Dump)  Data2(JX,JY,JV) = 1.0 - E
                     if (E .lt. Emin)  then
                        Emin = E
                        Mmin = Mass
                        do k=1,5
                           Xmin(k) = Xk(k)
                           end do
                        end if
                     if (E .lt. E2min)  then
                        E2min = E
                        M2min = Mass
                        Contr2 = Contr0
                        do k=1,5
                           X2min(k) = Xk(k)
                           end do
                        end if
                     end if 
                    end do
                  end do
               end do
c  Dumping the image of correlation coefficients:
            if (Dump)  then
               No_BUNIT = .false.
               BUNIT = 'rho'
               call Writing_output_image(Data2,NX_max,NY_max,NV_max,
     #         'rho'//No//'.fits', BlockSize, Naxes, BUNIT,No_BUNIT, 
     #         Type,Rval,Rpix,Delt,Rota,No_coord,Bitpix)

               No_BUNIT = .false.
               BUNIT = 'no_units'
               LXc = (Laxes(1)+1)/2
               LYc = (Laxes(2)+1)/2
               LVc = (Laxes(3)+1)/2
               do JX=1,Naxes(1)
                  LX = JX - Xmin(1) + LXc
                  do JY=1,Naxes(2)
                     LY = JY - Xmin(2) + LYc
                     do JV=1,Naxes(3)
                        LV = JV - Xmin(3) + LVc
                        if (LX.lt.1.or.LX.gt.Laxes(1) .or. 
     #                       LY.lt.1.or.LY.gt.Laxes(2) .or.
     #                       LV.lt.1.or.LV.gt.Laxes(3))  then
                           Data2(JX,JY,JV) = 0.0
                          else
                           Data2(JX,JY,JV) = LModel(LX,LY,LV)
                          end if
                        end do
                     end do
                  end do
               call Writing_output_image(Data2,NX_max,NY_max,NV_max,
     #         'mod'//No//'.fits', BlockSize, Naxes, BUNIT,No_BUNIT, 
     #         Type,Rval,Rpix,Delt,Rota,No_coord,Bitpix)
               end if

               IR0 = IR-IR1+1
               IV0 = IV-IV1+1
               Edat(1,IR0,IV0) = 1-E2min
               Edat(2,IR0,IV0) = Nt
               Edat(3,IR0,IV0) = X2min(1)
               Edat(4,IR0,IV0) = X2min(2)
               Edat(5,IR0,IV0) = X2min(3)
               Edat(6,IR0,IV0) = M2min
               Contr(IR0,IV0) = Contr2
               
            end do  !  end of IV
         end do   !  end of IR

      write (*,*) ' Number of steps:', MMM

      return
      end
c-------------------------------------------------------------------------


      function  Cost_function (Xk, Data,NX_max,NY_max,NV_max,
     #           Naxes, LModel,LX_max,LY_max,LV_max,Laxes,
     #           St,St2,Nt,Center,CM,Mass,Contrast)
c  2.081: Spheric boundaries.
c  2.11 : Interpolation between pixels.
c  2.12 : New interpolation algorithm.
c  3.29 : Works correctly with undefined Data pixels.
      implicit NONE
      integer NX_max,NY_max,NV_max, LX_max,LY_max,LV_max
      real Data(NX_max,NY_max,NV_max), LModel(LX_max,LY_max,LV_max),
     #     Cost_function, R, D,Contrast,
     #     Sd,Sd2,Std,E, 
     #     St,St2, Data0, Xk(3), Center(5),CM,Mass
      integer Naxes(4), Laxes(4), LX,LY,LV, IX,IY,IV, Nt, ID,IData0
      equivalence (D,ID), (Data0,IData0)


      Std = 0.0
      Sd = 0.0
      Sd2 = 0.0

      do LX=1,Laxes(1)
         IX = int(Xk(1)) + LX - (Laxes(1)+1)/2
         do LY=1,Laxes(2)
            IY = int(Xk(2)) + LY - (Laxes(2)+1)/2
            do LV=1,Laxes(3)
               IV = int(Xk(3)) + LV - (Laxes(3)+1)/2
               D = LModel(LX,LY,LV)               
               Data0 = Data(IX,IY,IV)
               if (ID.ne.-1 .and. IData0.ne.-1)  then
                  Std = Std + D*Data0
                  Sd = Sd + Data0
                  Sd2 = Sd2 + Data0**2
                  end if
               end do
            end do
         end do

      R = Sd2 - Sd**2/Nt
      if (R .ge. 0.0)  then
         E = 1.0 - (Std - St*Sd/Nt) / sqrt((St2 - St**2/Nt)*R)
        else
         E = 1.0
        end if
      Cost_function = E

      Mass = (Nt*Std - St*Sd) / CM
c  Contrast of the shell against the residual contant background:
      Contrast = abs(Mass) / (abs(Mass) + Nt*abs((St*Std-St2*Sd)/
     #           (St**2-Nt*St2)))

      return
      end
c-------------------------------------------------------------------------


      subroutine Reading_input_image (Name,Naxis,Naxes,No_BUNIT,BUNIT,
     #         Type,Rval,Rpix,Delt,Rota,No_coord,
     #         Array,NXmax,NYmax,NZmax, BlockSize,Bitpix, SFRAD,SFVEL)

      implicit NONE
      character*(*) Name, Comment*1, BUNIT, Type(3), No*1
      integer BlockSize, Status, Bitpix, Naxis, Naxes(4), Pcount,Gcount,
     #        NXmax,NYmax,NZmax, I,IX,IY,IV, Blank
      real*8 Rval(3)
      real Rpix(3), Delt(3), Rota(3), Array(NXmax,NYmax,NZmax), Nullval,
     #     SFRAD,SFVEL
      logical Simple, Extend, No_BUNIT, No_coord, Anyf
      equivalence (Nullval,Blank)
     
      No_BUNIT = .false.
      No_coord = .false.
      Status = 0
      
      call FTopen (1, Name, 0, BlockSize, Status)
      if (Status .ne. 0) then
         write (*,*) ' !!! Mistake in opening the file ', Name
         stop
         end if

      call FTghpr (1, 3, Simple, Bitpix, Naxis, Naxes, Pcount, Gcount,
     #            Extend, Status)
      if (Status .ne. 0)  then
         write (*,*) ' !!! Mistake in reading the header in ', Name
         stop
         end if

      if (Naxes(1).gt.NXmax.or.Naxes(2).gt.NYmax.or.Naxes(3).gt.NZmax)
     #     then
         write (*,*) ' Input image is bigger than',NXmax,'x',NYmax,'x',
     #               NZmax,' !!!'
         stop
         end if

      call FTgkyS (1, 'BUNIT', BUNIT, Comment, Status)
      if (Status .ne. 0)  then
         Status = 0
         No_BUNIT = .true.
         end if

      do I=1,3
         write (No,'(I1)') I
         call FTgkyS (1, 'CTYPE'//No//'  ', Type(I), Comment, Status)
         call FTgkyD (1, 'CRVAL'//No//'  ', Rval(I), Comment, Status)
         call FTgkyE (1, 'CDELT'//No//'  ', Delt(I), Comment, Status)
         call FTgkyE (1, 'CRPIX'//No//'  ', Rpix(I), Comment, Status)
c         call FTgkyE (1, 'CROTA'//No//'  ', Rota(I), Comment, Status)
         end do
      if (Status .ne. 0)  then
         Status = 0
         No_coord = .true.
         end if
      call FTgkyE (1, 'SFRAD', SFRAD, Comment, Status)
      call FTgkyE (1, 'SFVEL', SFVEL, Comment, Status)
      if (Status .ne. 0)  Status = 0

      do IX=1,NXmax
         do IY=1,NYmax
            do IV=1,NZmax
               Array(IX,IY,IV) = 0.0
               end do
            end do
         end do

c  "Magic blanking" of undefined pixels: we use a Real representation
c  of Integer -1 as a magic value:
      Blank = -1

c  Reading the data array into Array:
      call FTg3dE (1,0,Nullval, NXmax,NYmax, Naxes(1),Naxes(2),Naxes(3), 
     #             Array, Anyf, Status)

      if (Anyf)  then
         write (*,*) ' !!! Undefined values in the image !'
         end if
      if (Status .ne. 0)  then
         write (*,*) ' !!! Mistake in reading the file !'
         stop
         end if

      call FTclos (1, Status)

      return
      end
c-------------------------------------------------------------------------



      subroutine  Generate_model (Model,MX_max,MY_max,MV_max,Maxes,
     #            LModel,LX_max,LY_max,LV_max,Laxes,
     #            MXY,MV,St,St2,Nt,CM)

c  The most time consuming subroutine!
c  2.081: spheric boundaries are introduced.

      implicit NONE
      integer MX_max,MY_max,MV_max, IX,IY,IV, Maxes(3), Laxes(3), LX,LY,
     #        LV, KX1,KX2,KY1,KY2,KV1,KV2, LX_max,LY_max,LV_max,
     #        LXc,LYc,LVc, IBlank, IMag, ID,Nt
      real Model(MX_max,MY_max,MV_max), MXY,MV, IX1,
     #     IX2,IY1,IY2,IV1,IV2, dX1,dX2,dX,dY1,dY2,dY,dV1,dV2,dV,
     #     aX,aY,aV, LModel(LX_max,LY_max,LV_max), Blank,Mag,D,St,St2,CM
      logical LBlank
      equivalence (Blank,IBlank), (IMag,Mag), (ID,D)

      IMag = -1

      LXc = (Laxes(1)+1)/2
      LYc = (Laxes(2)+1)/2
      LVc = (Laxes(3)+1)/2

c  Scaling 1/8 part of the model:

      do LX=LXc,Laxes(1)
         IX1 = (LX-LXc)/MXY + 0.5
         IX2 = (LX-LXc+1)/MXY + 0.5
         if (IX1 .lt. 0.0)  IX1 = 0.0
         if (IX2 .gt. Maxes(1)-1.0)  IX2 = Maxes(1)-1.0
         dX1 = 1 - IX1 + int(IX1)
         dX2 = IX2 - int(IX2)
         KX1 = int(IX1) + 1
         KX2 = int(IX2) + 1
         do LY=LYc,Laxes(2)
            IY1 = (LY-LYc)/MXY + 0.5
            IY2 = (LY-LYc+1)/MXY + 0.5
            if (IY1 .lt. 0.0)  IY1 = 0.0
            if (IY2 .gt. Maxes(2)-1.0)  IY2 = Maxes(2)-1.0
            dY1 = 1 - IY1 + int(IY1)
            dY2 = IY2 - int(IY2)
            KY1 = int(IY1) + 1
            KY2 = int(IY2) + 1
            do LV=LVc,Laxes(3)
               IV1 = (LV-LVc)/MV + 0.5
               IV2 = (LV-LVc+1)/MV + 0.5
               if (IV1 .lt. 0.0)  IV1 = 0.0
               if (IV2 .gt. Maxes(3)-1.0)  IV2 = Maxes(3)-1.0
               dV1 = 1 - IV1 + int(IV1)
               dV2 = IV2 - int(IV2)
               KV1 = int(IV1) + 1
               KV2 = int(IV2) + 1

               LModel(LX,LY,LV) = 0.0
               LBlank = .false.
               do IX=KX1,KX2
                  if (IX .ne. 1)  then
                     aX = 1.0
                    else                     
                     aX = 2.0
                    end if
                  if (IX .eq. KX1)  then
                     dX = dX1
                    else
                     if (IX .eq. KX2)  then
                        dX = dX2
                       else
                        dX = 1.0
                       end if
                    end if
                  do IY=KY1,KY2
                     if (IY .ne. 1)  then
                        aY = 1.0
                       else
                        aY = 2.0
                       end if
                     if (IY .eq. KY1)  then
                        dY = dY1
                       else
                        if (IY .eq. KY2)  then
                           dY = dY2
                          else
                           dY = 1.0
                          end if
                       end if
                     do IV=KV1,KV2
                        if (IV .ne. 1)  then
                           aV = 1.0
                          else
                           aV = 2.0
                          end if
                        if (IV .eq. KV1)  then
                           dV = dV1
                          else
                           if (IV .eq. KV2)  then
                              dV = dV2
                             else
                              dV = 1.0
                             end if
                          end if
                        Blank = Model(IX,IY,IV)
                        if (IBlank .ne. -1)  then
                           LModel(LX,LY,LV) = LModel(LX,LY,LV) +
     #                          dX*dY*dV * aX*aY*aV * Blank
                          else
                           LBlank = .true.   
                          end if
                        end do
                     end do
                  end do
               if (LBlank .and. LModel(LX,LY,LV).eq.0.0) 
     #              LModel(LX,LY,LV)=Mag
c  Filling the rest of the model using symmetries:
               LModel(2*LXc-LX,LY,LV) = LModel(LX,LY,LV)
               LModel(LX,2*LYc-LY,LV) = LModel(LX,LY,LV)
               LModel(LX,LY,2*LVc-LV) = LModel(LX,LY,LV)
               LModel(2*LXc-LX,2*LYc-LY,LV) = LModel(LX,LY,LV)
               LModel(LX,2*LYc-LY,2*LVc-LV) = LModel(LX,LY,LV)
               LModel(2*LXc-LX,LY,2*LVc-LV) = LModel(LX,LY,LV)
               LModel(2*LXc-LX,2*LYc-LY,2*LVc-LV) = LModel(LX,LY,LV)

               end do
            end do
         end do

c !!!! Fixed bug:
      Nt = 0
      St = 0.0
      St2 = 0.0
      do LX=1,Laxes(1)
         do LY=1,Laxes(2)
            do LV=1,Laxes(3)
               D = LModel(LX,LY,LV)
               if (ID .ne. -1)  then
                  Nt = Nt + 1
                  St = St + D
                  St2 = St2 + D**2
                  end if
               end do
            end do
         end do

c  For mass estimation:
      CM = (Nt*St2-St**2) / ST

      return
      end
c-------------------------------------------------------------------------


      subroutine Calculate_limits (R,V,SFRAD,SFVEL,
     #                           Naxes,Maxes, Laxes,Limits,MXY,MV,Wrong,
     #                           Center,Frac)
      implicit NONE
      integer Naxes(3),Maxes(3),Laxes(3), I
      real R,V,SFRAD,SFVEL, MXY,MV,Limits(2,5),
     #     Size(3),Center(3),Frac(3)
      logical Wrong
      
      Wrong = .false.

      MXY = R / SFRAD
      MV = V / SFVEL
      if (MXY.gt.1.0 .or. MV.gt.1.0)  then
         Wrong = .true.
         return
         end if

      Laxes(1) = MXY * (2*Maxes(1)+1)
      Laxes(2) = MXY * (2*Maxes(2)+1)
      Laxes(3) = MV * (2*Maxes(3)+1)
      Size(1) = R
      Size(2) = R
      Size(3) = V
c  It should be odd:
      if (mod(Laxes(1),2) .eq. 0)  Laxes(1)=Laxes(1)+1
      if (mod(Laxes(2),2) .eq. 0)  Laxes(2)=Laxes(2)+1
      if (mod(Laxes(3),2) .eq. 0)  Laxes(3)=Laxes(3)+1

      do I=1,3         
c !!!! Fixed bug with Laxes(1):
         Limits(1,I) = -amin1(Center(I)-(Laxes(I)+1)/2,
     #      anint(Frac(I)*Size(I)))/Size(I)
         Limits(2,I) =  amin1(Naxes(I)-Center(I)-(Laxes(I)-1)/2,
     #      anint(Frac(I)*Size(I)))/Size(I)
         end do

      return
      end
c-------------------------------------------------------------------------


      subroutine Writing_output_image(Out,NXOmax,NYOmax,NZOmax,
     #         OName, BlockSize, Naxes, BUNIT,No_BUNIT, Type,Rval,
     #         Rpix,Delt,Rota,No_coord,Bitpix)

      implicit NONE
      real Data_max,Data_min
      integer IX,IY,IZ,NXOmax,NYOmax,NZOmax,BlockSize,Naxes(4),
     #        Status, Bitpix, I, ID
      real Out(NXOmax,NYOmax,NZOmax), Rpix(3), Delt(3), Rota(3),D
      real*8 Rval(3)
      character*(*) OName, BUNIT, Type(3), No*1
      logical Simple, Extend, No_BUNIT, No_coord
      equivalence (D,ID)

      ID = -1

      Status = 0
c  Finding the extreme values:
      Data_max = -1.e33   
      Data_min = 1.e33
      do IX=1,Naxes(1)
         do IY=1,Naxes(2)
            do IZ=1,Naxes(3)
               D = Out(IX,IY,IZ)
               if (ID .ne. -1)  then
                  if (Out(IX,IY,IZ).gt.Data_max)  Data_max=Out(IX,IY,IZ)
                  if (Out(IX,IY,IZ).lt.Data_min)  Data_min=Out(IX,IY,IZ)
                  end if
               end do
            end do
         end do

c  Writing the output FITS file:
      open (1, file=OName)
      close (1, status='DELETE')
      call FTinit (1, OName,Blocksize, Status)

      if (Status .ne. 0)  then
         write (*,*) '!!! Cannot create file', OName
         write (*,*) ' Status = ', Status
         stop
         end if

      Simple = .true.
      Extend = .false. 
c  Writing the primary keywords:
      call FTphpr (1, Simple, -32, 3, Naxes, 0, 1, Extend, Status)   
      if (Status .ne. 0)  stop

c  Writing the additional keywords:
      call FTpkyL (1, 'BLOCKED ', .true., ' ', Status)
      if (.not. No_BUNIT)  then
         call FTpkyS (1, 'BUNIT   ', BUNIT, ' ', Status)
         end if
      call FTpkyE (1, 'DATAMAX ', Data_max, 10, ' ', Status)
      call FTpkyE (1, 'DATAMIN ', Data_min, 10, ' ', Status)

      if (.not. No_coord)  then
          do I=1,3
             write (No,'(I1)') I
             call FTpkyS (1, 'CTYPE'//No//'  ', Type(I), ' ', Status)
             call FTpkyD (1, 'CRVAL'//No//'  ', Rval(I), 10, ' ',Status)
             call FTpkyE (1, 'CDELT'//No//'  ', Delt(I), 10, ' ',Status)
             call FTpkyE (1, 'CRPIX'//No//'  ', Rpix(I), 10, ' ',Status)
c             call FTpkyE (1, 'CROTA'//No//'  ', Rota(I), 10, ' ',Status)
             end do
          end if
         
      call FTpdat (1, Status)

      if (Status .ne. 0)  then
         write (*,*) '!!! Wrong header parameter'
         stop
         end if

c Writing the image array:

      call FTp3dE (1,1, NXOmax,NYOmax, Naxes(1),Naxes(2),Naxes(3),
     #            Out, Status)

      if (Status .ne. 0)  then
         write (*,*) '!!! Error while writing the data array'
         stop
         end if

      call FTclos (1, Status)

      return
      end
c-------------------------------------------------------------------------


      INTEGER FUNCTION strlen_lead (STRING)
C
C Return the length of a string less any leading blanks
C
C
C   STRING      CH*(*)  input   String to be measured

      implicit NONE
      CHARACTER*(*)     STRING
      INTEGER           I, N

      N = LEN (STRING)
      DO 1 I = 1,N
         IF(STRING(I:I).NE.' ') GOTO 2
    1 CONTINUE
    2 STRLEN_lead = N-I+1
      return
      END
c-------------------------------------------------------------------------
      SUBROUTINE gaussj(a,n,np,b,m,mp)
      INTEGER m,mp,n,np,NMAX
      REAL a(np,np),b(np,mp)
      PARAMETER (NMAX=50)
      INTEGER i,icol,irow,j,k,l,ll,indxc(NMAX),indxr(NMAX),ipiv(NMAX)
      REAL big,dum,pivinv
      do 11 j=1,n
        ipiv(j)=0
11    continue
      do 22 i=1,n
        big=0.
        do 13 j=1,n
          if(ipiv(j).ne.1)then
            do 12 k=1,n
              if (ipiv(k).eq.0) then
                if (abs(a(j,k)).ge.big)then
                  big=abs(a(j,k))
                  irow=j
                  icol=k
                endif
              else if (ipiv(k).gt.1) then
                pause 'singular matrix in gaussj'
              endif
12          continue
          endif
13      continue
        ipiv(icol)=ipiv(icol)+1
        if (irow.ne.icol) then
          do 14 l=1,n
            dum=a(irow,l)
            a(irow,l)=a(icol,l)
            a(icol,l)=dum
14        continue
          do 15 l=1,m
            dum=b(irow,l)
            b(irow,l)=b(icol,l)
            b(icol,l)=dum
15        continue
        endif
        indxr(i)=irow
        indxc(i)=icol
        if (a(icol,icol).eq.0.) pause 'singular matrix in gaussj'
        pivinv=1./a(icol,icol)
        a(icol,icol)=1.
        do 16 l=1,n
          a(icol,l)=a(icol,l)*pivinv
16      continue
        do 17 l=1,m
          b(icol,l)=b(icol,l)*pivinv
17      continue
        do 21 ll=1,n
          if(ll.ne.icol)then
            dum=a(ll,icol)
            a(ll,icol)=0.
            do 18 l=1,n
              a(ll,l)=a(ll,l)-a(icol,l)*dum
18          continue
            do 19 l=1,m
              b(ll,l)=b(ll,l)-b(icol,l)*dum
19          continue
          endif
21      continue
22    continue
      do 24 l=n,1,-1
        if(indxr(l).ne.indxc(l))then
          do 23 k=1,n
            dum=a(k,indxr(l))
            a(k,indxr(l))=a(k,indxc(l))
            a(k,indxc(l))=dum
23        continue
        endif
24    continue
      return
      END
c-------------------------------------------------------------------------


      INTEGER FUNCTION strlen_trail (STRING)
C
C Return the length of a string less any trailing blanks
C
C
C   STRING      CH*(*)  input   String to be measured
      implicit NONE
      CHARACTER*(*)     STRING
      INTEGER           I, N

      N = LEN (STRING)
      DO 1 I = N,1,-1
         IF(STRING(I:I).NE.' ') GOTO 2
    1 CONTINUE
    2 STRLEN_trail = I
      END
c--------------------------------------------------------------------------
